#!/usr/bin/python3

import argparse
import hashlib
import json
import os
import pprint
import re
import ssl
import sys

sys.path.append("debian/lib/python")

import debian_linux.config
import debian_linux.debian

parser = argparse.ArgumentParser(
    description="Generate files for signing-template package.")
parser.add_argument(
    "--featureset",
    help="Process only kernel package with FEATURESET. Default: all",
)
parser.add_argument(
    "arch",
    help="Architecture for which packages are being built.",
)
args = parser.parse_args()

arch = args.arch
assert arch == "amd64", arch

# Create output directory structure
template_top_dir = "debian/linux-image-%(arch)s-signed-template/usr/share/code-signing/linux-image-%(arch)s-signed-template" % {
    "arch": arch
}
template_debian_dir = template_top_dir + "/source-template/debian"
os.makedirs(template_debian_dir, exist_ok=True)

# Initialize template variables
changelog = debian_linux.debian.Changelog(
    version=debian_linux.debian.VersionLinux)
version = changelog[0].version

abiname_version = re.sub("^(\d+\.\d+)(?=-|$)", r"\1.0", version.linux_upstream)

config = debian_linux.config.ConfigCoreHierarchy({}, ["debian/config"])

if version.linux_modifier is not None:
    abiname_part = ""
else:
    abiname_part = "-%s" % config["abi",]["abiname"]

template_vars = {
    "arch": arch,
    "abiname": abiname_version + abiname_part,
    "version": version.linux_version,
    "imagebinaryversion": version.complete,
    "image-stem": "vmlinuz",
}

# Generate changelog
with open(template_debian_dir + "/changelog", "w", encoding="utf-8") as f_out, \
        open("debian/changelog", "r", encoding="utf-8") as f_in:
    line = f_in.readline()
    _, _, rest = line.split(maxsplit=2)
    native_version = template_vars["imagebinaryversion"].replace("-", "+")
    f_out.write(
        """\
linux-signed-%s (%s) %s

  * Sign kernel from %s %s
""" % (
            arch, native_version, rest.strip(), changelog[0].source,
            version.complete))

    for d in f_in.read():
        f_out.write(d)


def substitute(s, vars):
    if isinstance(s, (list, tuple)):
        return [substitute(i, vars) for i in s]

    def subst(match):
        return vars[match.group(1)]

    return re.sub(r"@([-_a-z0-9]+)@", subst, s)


def substitute_file(template, target, vars, append=False):
    with open(target, "a" if append else "w", encoding="utf-8") as f:
        f.write(substitute(template, vars))


# Some functions in openssl work with multiple concatenated PEM-format
# certificates, but others do not.
def get_certs(file_name):
    certs = []
    BEGIN, MIDDLE = 0, 1
    state = BEGIN
    with open(file_name) as f:
        for line in f:
            if line == "-----BEGIN CERTIFICATE-----\n":
                assert state == BEGIN
                certs.append([])
                state = MIDDLE
            elif line == "-----END CERTIFICATE-----\n":
                assert state == MIDDLE
                state = BEGIN
            else:
                assert line[0] != "-"
                assert state == MIDDLE
            certs[-1].append(line)
    assert state == BEGIN
    return ["".join(cert_lines) for cert_lines in certs]


def get_cert_fingerprint(cert, algo):
    hasher = hashlib.new(algo)
    hasher.update(ssl.PEM_cert_to_DER_cert(cert))
    return hasher.hexdigest()

all_files = {"packages": {}}

build_depends = []
control_binary = []
recipes = {}

featuresets = config["base",]["featuresets"]
if args.featureset is not None:
    assert args.featureset in featuresets, (args.featureset, featuresets)
    featuresets = [args.featureset]
for featureset in featuresets:
    if not config.merge("base", None, featureset).get("enabled", True):
        continue

    featureversion = ""
    if featureset != "none":
        featureversion = "-" + featureset

    template_vars["featureset"] = featureset

    for flavour in config["base", arch, featureset]["flavours"]:
        template_vars["localversion"] = "%s-%s" % (featureversion, flavour)

        pprint.pprint(template_vars)

        # Generate maintainer scripts
        for name in ["postinst", "postrm", "preinst", "prerm", "templates"]:
            substitute_file(
                open("debian/templates/image.plain.%s.in" % (name,)).read(),
                template_debian_dir + "/linux-image-%s%s.%s" % (
                    template_vars["abiname"],
                    template_vars["localversion"],
                    name,
                ),
                template_vars,
            )

        # Collect data for files.json
        package_files = []
        image_suffix = "%(abiname)s%(localversion)s" % template_vars
        package_files.append({
            "sig_type": "efi",
            "file": "boot/vmlinuz-%s" % image_suffix
        })
        image_package_name = "linux-image-%s-unsigned" % (image_suffix,)
        package_dir = "debian/%s" % (image_package_name,)

        kconfig = open("%s/boot/config-%s" %
                       (package_dir, image_suffix)).readlines()
        cert_re = re.compile("CONFIG_SYSTEM_TRUSTED_KEYS=\"(.*)\"$")
        cert_file_name = None
        for line in kconfig:
            match = cert_re.match(line)
            if match:
                cert_file_name = match.group(1)
                break
        assert cert_file_name
        if featureset != "none":
            cert_file_name = os.path.join(
                "debian/build/source_%s" % (featureset,), cert_file_name)

        package_certs = [
            get_cert_fingerprint(cert, "sha256")
            for cert in get_certs(cert_file_name)
        ]

        def raise_func(e):
            raise e

        package_modules = []
        for root, dirs, files in os.walk("%s/lib/modules" % package_dir,
                                         onerror=raise_func):
            package_modules.extend([
                "%s/%s" % (root[(len(package_dir) + 1):], name)
                for name in files
                if name.endswith(".ko")
            ])
        package_modules.sort()
        package_files.extend([{
            "sig_type": "linux-module",
            "file": module,
        } for module in package_modules])
        all_files["packages"][image_package_name] = {
            "trusted_certs": package_certs,
            "files": package_files
        }

        # Collect data for control
        control_binary.append(
            substitute(
                open("debian/signing_templates/control.image.in").read(),
                template_vars))
        build_depends.append("%s [%s]" % (image_package_name, arch))

        target = "binary-arch_%s_%s" % (arch, featureset)
        recipes[target] = substitute(
            open("debian/signing_templates/rules.gen.in").read(), template_vars)

# Generate files.json
with open(template_top_dir + "/files.json", "w", encoding="utf-8") as f:
    json.dump(all_files, f, indent="  ")
    f.write("\n")

# Generate control
template_vars["image_packages"] = ", ".join(build_depends)
with open(template_debian_dir + "/control", "w", encoding="utf-8") as f:
    f.write(
        substitute(
            open("debian/signing_templates/control.source.in").read(),
            template_vars))
    for package in control_binary:
        f.write(package)

# Generate rules.gen
with open(template_debian_dir + "/rules.gen", "w", encoding="utf-8") as f:
    k, v = zip(*recipes.items())
    f.write("""\
.NOTPARALLEL:
binary-arch_amd64:: %s
""" % (" ".join(k)))
    f.write("".join(v))
